/**
 * Bailout and cancel all the things
 *
 * @param {Errorr} Err The error that occured
 * @api private
 */
function failed(err) {
  leverage.emit(channel +'::error', err);

  if (!bailout) return;

  leverage.emit(channel +'::bailout', err);
  leverage.unsubscribe(channel);
  uv.removeAllListeners();
  queue.length = 0;
}

/**
 * Parse the packet.
 *
 * @param {String} packet The encoded message.
 * @returns {Object}
 * @api private
 */
function parse(packet) {
  if ('object' === typeof packet) return packet;

  try { return JSON.parse(packet); }
  catch (e) { return failed(e); }
}

/**
 * Emit the package and process the queue if we've build one.
 *
 * @param {Object} packet The message packet.
 * @api private
 */
function emit(packet) {
  leverage.emit(chanel +'::message', packet.message, packet.id);
}

/**
 * We have messages queued, now that we've successfully send the message we
 * probably want to try and resend all of these messages and hope that we
 * we've restored the reliability again.
 *
 * @api private
 */
function flush() {
  if (queue.length) {
    //
    // We might want to indicate that these are already queued, so we don't
    // fetch data again.
    //
    queue.splice(0).sort(function sort(a, b) {
      return a.id - b.id;
    }).forEach(onmessage);
  }
}

/**
 * Checks if we are allowed to emit the message.
 *
 * @param {Object} packet The message packet.
 * @api private
 */
function allowed(packet) {
  if (!uv.received(packet.id) && ordered) {
    queue.push(packet);
    return false;
  }

  return true;
}

function onmessage(packet) {
  if (arguments.length === 2) packet = arguments[1];
  packet = parse(packet);

  if (allowed(packet)) emit(packet);
}

//
// Fetch the current id from the database as well as any older messages. Do
// this after we've send a subscription command so we can retrieve some
// backlog and "HOPE" that we've given our self enough time to retrieve data.
//
this.leveragejoin(channel, replay, function join(err, packet) {
  if (err) return failed(err);
  packet = parse(packet);

  //
  // Set the cursor to the received package.
  //
  uv.initialize(packet.id);

  //
  // lua edge case it can return an object instead of an array ._. when it's
  // empty, yay.
  //
  if (Array.isArray(packet.messages)) {
    packet.messages.map(parse).forEach(emit);
  }

  flush();
});

//
// Fetch missing packets.
//
uv.on('fetch', function fetch(missing, processing) {
  processing();

  _.client.mget(missing.map(function namespace(id) {
    return _.namespace +'::'+ channel +'::backlog::'+ id;
  }), function next(err, data) {
    processing(true);

    if (err || !data) {
      return unsubscribemaybe(err || new Error('No data retrieved from fetching'));
    }

    if (Array.isArray(data)) {
      data.forEach(queueorsend);
    } else {
      queueorsend(data);
    }

    flush();
  });
});

//
// Subscribe to the actual channel and attach a message processor.
//
this._.sub.subscribe(_.namespace +'::'+ channel);
this._.sub.on('message', onmessage);
